package cam.netapp;
import java.io.*;
import java.net.*;
//
//##################################################################
//  		       Class NetworkConnection
//##################################################################
//
//	            Originator    : Dan Delsol (Winter 97')
//
//              Modifications : Chris Anderson 4/28/97, 9/1/98 present
//              Modifications : Dan Delsol 5/07/97,5/16/97
//              (C) UCLA
//
//##################################################################
//
// Design Notes : I'm using String communication to keep the
// implementation simple and easy to debug.
//
/**
An instance of this class is typically associated with the &quot;front end&quot; of a distributed application.
The NetworkConnection class performs two important functions</P>
<UL>
	<LI>It instantiates a remote class and sets up stream communications with that instance. It does this by communicating
	with a cam.netapp.ServerManager executing on the remote machine.
	<LI>It provides run-time control of the remote class instance.
</UL>
<P>The &quot;back-end&quot; does not require an instance of this class (it is only required to implement the Runnable
interface and the SetStreams interface).</P>
<P>The remote class usually resides on another machine, but for development purposes, one can use a NetworkConnection
to connect to a class on a local machine.</P>
<P>In order to establish a remote class instance on another machine, an instance of the ServerManager class must
be executing on that machine.</P>
@author Chris Anderson and Dan Delsol (C) UCLA 1997-98
@version 9/18/98
*/
public class NetworkConnection
{
/**
Standard constructor.
*/
public NetworkConnection()
{
    sockets        = null;
    socketCount    = 2;

    controlIn      = null;
    controlOut     = null;
    DataIstream    = null;
    ControlIstream = null;
    DataOstream    = null;
    ControlOstream = null;

    verbose        = false;
}// constructor

/**
Creates a network connection using the sockets passed in sArray for
communication. The control streams are <b>NOT</b> created with this construction.
Control commands (startRemoteApplication(), stopRemoteApplication()... have
no effect.

If the control streams are desired, then the createControlStreams() method
should be called after the constructor. Use of the control streams presumes
that there is a remote ControlThread instance connected to sockets[1].
*/
public NetworkConnection(Socket [] sArray)
{


    sockets     = sArray;
    socketCount = sockets.length;

    controlIn      = null;
    controlOut     = null;
    DataIstream    = null;
    ControlIstream = null;
    DataOstream    = null;
    ControlOstream = null;

    verbose     = false;
}
/**
Creates a connection to the application appName at (address,port).
@param address Address of remote machine upon which a ServerManager is executing
@param portnum Port where the remote machine that the ServerManager is listening. (Default = 6789).
@param appName Name of the class on the remote machine
@exception cam.netapp.NetAppException if the connection can not be completed.
*/
public void connectRemoteApplication(String address, int portnum, String appName)
throws NetAppException
{
//
// address : address of remote machine
// portnum : listening port for the remote ServerManager (fixed)
// appName : name of remote application
//
   BufferedReader    sin;   // stream wrappers for startup communication
   PrintWriter      sout;
//
// Set up connections : a three step process
//
// (i)   connect to server manager; get port for JavaServer
// (ii)  connect to JavaServer and set up socket connections
// (iii) request application from JavaServer
//
   try
   {
//
//  Connect to the ServerManager using the public port
//  and then read back an available remote port upon which the
//  JavaServer application will be listening.
//
    if(verbose)
    {
    System.out.println("NetworkConnection : Setting up Connection");
    System.out.println("NetworkConnection : Connecting to ["
                              + address + "] Port : " + portnum);
    }

    startUp        = new Socket(address, portnum);
    sin            = new BufferedReader(new InputStreamReader(startUp.getInputStream()));
    int remotePort = (new Integer(sin.readLine())).intValue();
//
//  Connect to the JavaServer using the available remote port
//
    startUp        = new Socket(address, remotePort);
    sin            = new BufferedReader(new InputStreamReader(startUp.getInputStream()));
    sout           = new PrintWriter(startUp.getOutputStream(),true);

    if(verbose)
    {
    System.out.println("NetworkConnection : Connected  to [" + address + "] Start Up Port : "
                                                         + remotePort);
    }
//
//  allocate an array of sockets
//
    sockets = new Socket[socketCount];
//
//  send number of socket connections to make
//
    sout.println(new Integer(socketCount));
    sout.flush();

    int k;
    for(k = 0; k < socketCount; k++)
    {
//
//  Read port number for socket
//
    remotePort = (new Integer(sin.readLine())).intValue();
//
//  Create socket connection
//
    sockets[k]      = new Socket(address, remotePort);
    if(verbose)
    {
    System.out.println("NetworkConnection : Connected  to [" + address + "] Server   Port : "
                                                         + remotePort);
    }
    }// k loop
//
//  Write the name of the application which will be dynamically loaded
//  and set running. Catch any errors in the process...and throw an exception
//  if required.
//
    if(verbose)
    {
    System.out.println("NetworkConnection : Application " + appName + " requested");
    }

    sout.println(appName);
    String S1 = sin.readLine();
    if(verbose)System.out.println("NetworkConnection : "+S1);
    if(!(S1.equals("Remote Class Found")))
    {
    NetAppException ee = new NetAppException(S1);
    throw ee;
    };

    String S2 = sin.readLine();
    if(verbose)System.out.println("NetworkConnection : "+S2);
    if(!(S2.equals("Remote Class Instantiated")))
    {
    NetAppException ee = new NetAppException(S2);
    throw ee;
    };

    String S3 = sin.readLine();
    if(verbose)System.out.println("NetworkConnection : "+S3);
    if(!(S3.equals("Streams Connected to Remote Class")))
    {
    NetAppException ee = new NetAppException(S3);
    throw ee;
    };

//
//  sockets connected to the remote application
//
    }//try
    catch(IOException e)
    {
    NetAppException ee = new NetAppException("Problem Connecting With Remote Host/Application");
    throw ee;
    }
    catch(Exception e){NetAppException ee = new NetAppException(e.getMessage());
    throw ee;}


    if(verbose){System.out.println("NetworkConnection : Connection Complete");};
//
//  create control streams
//
    createControlStreams();
//
// clean up
//
    try
    {
    sin.close();
    sout.close();
    startUp.close();
    }
    catch(Exception e){};
//
//
//
} // connectToRemoteApplication

/**
Creates a connection to a local class appName. An appName instance is loaded and
a cross-stream connection is made to it. A separate thread is then created for
the instance and it's run() method is invoked.
@param appName Name of the class on the remote machine
@exception cam.netapp.NetAppException if the connection can not be completed.
*/
public void connectRemoteApplication(String nameApp) throws NetAppException
{
   socketCount = 0;
   if(verbose)System.out.println("NetworkConnection : Setting up Local Connection");
//
// Create "cross-connected" input/output streams for the data and control
//
	PipedOutputStream  BackDataOstream = null;
	BackDataOstream  = new PipedOutputStream();
    try{DataIstream  = new PipedInputStream(BackDataOstream);}
    catch(IOException e){NetAppException ee = new NetAppException(e.getMessage()); throw ee;}

    PipedInputStream   BackDataIstream = null;
	DataOstream         = new PipedOutputStream();
    try{BackDataIstream = new PipedInputStream(DataOstream);}
    catch(IOException e){NetAppException ee = new NetAppException(e.getMessage()); throw ee;}

	PipedOutputStream  BackControlOstream = null;
	BackControlOstream  = new PipedOutputStream();
    try{ControlIstream  = new PipedInputStream(BackControlOstream);}
    catch(IOException e){NetAppException ee = new NetAppException(e.getMessage()); throw ee;}

    PipedInputStream   BackControlIstream = null;
	ControlOstream           = new PipedOutputStream();
    try{BackControlIstream   = new PipedInputStream(ControlOstream);}
    catch(IOException e){NetAppException ee = new NetAppException(e.getMessage()); throw ee;}
//
//  Load the local class and set things up
//
    if(verbose)
    {System.out.println("NetworkConnection : Loading Local Class " + nameApp);}

    Class C = null;
    try
    {
        C = Class.forName(nameApp);
    }// try
    catch(ClassNotFoundException e)
    {
        NetAppException ex = new NetAppException("The class " + nameApp + " was not found.");
        throw(ex);
    }// catch

    Object ObInstance = null;
    try
    {
        ObInstance= C.newInstance();
    }
    catch(Exception e)
    {
        ObInstance = null;
        NetAppException ex = new NetAppException("The class " + nameApp + " could not be instantiated");
        throw(ex);
    }

// Change Object into wanted object type
// and set streams to allow communication with the JavaClient

    if(ObInstance != null)
    {
     if(verbose)
     {System.out.println("NetworkConnection : Connecting streams to " + nameApp);}
     try
     {
     ((SetStreams)ObInstance).setInputStream(BackDataIstream);
     ((SetStreams)ObInstance).setOutputStream(BackDataOstream);
     }
     catch(Exception e)
     {
     NetAppException ex = new NetAppException("Streams could not be connected to the class " + nameApp);
     throw(ex);
     };


     ControlThread appControlThread = new ControlThread();
     appControlThread.setName(nameApp + " control");

     try{appControlThread.setInputStream(BackControlIstream);}
     catch(Exception e){NetAppException ee = new NetAppException(e.getMessage()); throw ee;}
     try{appControlThread.setOutputStream(BackControlOstream);}
     catch(Exception e){NetAppException ee = new NetAppException(e.getMessage()); throw ee;}

     Thread current = Thread.currentThread();
     appControlThread.setPriority(current.getPriority() - 2);

     appControlThread.setRunnableObject(ObInstance,nameApp);
     appControlThread.start();
     //
     // set up streams for control thread
     //
     try
     {
     controlOut = new PrintWriter(ControlOstream,true);
     controlIn  = new BufferedReader(new InputStreamReader(ControlIstream));
     }
     catch(Exception e){};
     if(verbose){System.out.println("NetworkConnection : Connection Complete");};
    }

} // connectToRemoteApplication

/**
Creates a connection to the application appName at (address,port) with auxConnections
auxillary socket connections. Typically auxConnections = 1 for use in creating
NetworkConnection callbacks from the application to the client.
@param address Address of remote machine upon which a ServerManager is executing
@param portnum Port where the remote machine that the ServerManager is listening. (Default = 6789).
@param appName Name of the class on the remote machine
@param auxConnections The number of auxillary connections to the remote class 
(typically 0, unless one is implementing callbacks).
@exception cam.netapp.NetAppException if the connection can not be completed.
*/
public void connectRemoteApplication(String address, int portnum, String appName, int auxConnections)
throws NetAppException
{

    socketCount    = 2 + auxConnections;
    connectRemoteApplication(address, portnum, appName);
}
/**
Stops the remote class thread and closes connections to the remote class.
*/
public void disconnectRemoteApplication()
{
    if(verbose){System.out.println("NetworkConnection : Disconnected");};
    if(controlOut == null) return;
    controlOut.println("kill");
    controlOut.flush();
}
/**
Creates a thread for the remote class and invokes the run() method of the remote 
class.
*/
public void startRemoteApplication()
{
    if(verbose){System.out.println("NetworkConnection : Starting Remote Application");};
    if(controlOut == null) return;
    controlOut.println("start");
    controlOut.flush();
}
/**
Stops the remote class thread.
*/
public void stopRemoteApplication()
{
    if(verbose)
    {System.out.println("NetworkConnection : Stopping Remote Application");};
    if(controlOut == null) return;
    controlOut.println("stop");
    controlOut.flush();
}
/**
Pauses the remote class thread.
*/
public void pauseRemoteApplication()
{
    if(verbose)
    {System.out.println("NetworkConnection : Pausing Remote Application");};
    if(controlOut == null) return;
    controlOut.println("pause");
    controlOut.flush();
}
/**
Resumes the remote class thread.
*/
public void resumeRemoteApplication()
{
    if(verbose)
    {System.out.println("NetworkConnection : Resuming Remote Application");};
    if(controlOut == null) return;
    controlOut.println("resume");
    controlOut.flush();
}
/**
Returns the status of the remote class. 
*/
public String getRemoteApplicationStatus()
{
    if(controlOut == null) return null;

    boolean errorFlag    = false;
    String returnString   = new String();
    controlOut.println("status");
    controlOut.flush();
    try
    {returnString = controlIn.readLine();}
    catch(IOException e){returnString = null;}
    return returnString;
}
/**
Returns the primary socket connected to the remote class.
*/
public Socket getSocket()
{
    if(socketCount == 0) return null;
    return sockets[0];
}
/**
Returns the control socket connected to the remote class's control.
*/
public Socket getControlSocket()
{
    if(socketCount == 0) return null;
    return sockets[1];
}
/**
Returns the the kth auxillary socket associated with the current connection.
@param k The index of the socket.
*/
public Socket getAuxSocket(int k)
{
    if(socketCount < 2 + k) return null;
    return sockets[1+k];
}
/**
Returns the OutputStream associated with the socket 
connection to the remote instance. One writes to this stream to send
data to the remote class.
*/
public OutputStream getOutputStream()
{
    if(socketCount == 0) return DataOstream;
    try{return sockets[0].getOutputStream();} catch(Exception e){};
    return null;
}
/**
Returns the OutputStream associated with the control socket 
connection. One writes to this stream to send
data to the remote class's control. 
*/
public OutputStream getControlOutputStream()
{
    if(socketCount == 0) return ControlOstream;
    try{return sockets[1].getOutputStream();} catch(Exception e){};
    return null;
}
/**
Returns the InputStream associated with the socket 
connection to the remote instance. One reads from this stream to obtain 
data from the remote class.
*/
public InputStream getInputStream()
{
    if(socketCount == 0) return DataIstream;
    try{return sockets[0].getInputStream();} catch (Exception e){};
    return null;
}
/**
Returns the InputStream associated with the control Socket connection.
One reads from this stream to obtain data from the remote class's control.
*/
public InputStream getControlInputStream()
{
    if(socketCount == 0) return ControlIstream;
    try{return sockets[1].getInputStream();} catch (Exception e){};
    return null;
}
/**
If set <b>true</b> diagnostic messages are printed to System.out.
*/
public void setVerboseFlag(boolean v)
{
    verbose = v;
}
/**
Create streams to communicate with the control thread using sockets[1]
*/
public void createControlStreams()
{

    if(socketCount > 1)
    {
    try{
    controlIn           = new BufferedReader(new InputStreamReader(sockets[1].getInputStream()));
    controlOut          = new PrintWriter(sockets[1].getOutputStream(),true);
    }
    catch(Exception e){};
    }
}
/**
Stops the remote class thread.
*/
protected void finalize()
{
    controlOut.println("kill");
}
//
// members
//
   private Socket [] sockets; // For communication with the application
                              //
                              // sockets[0]    = socket for primary communication with client
                              // sockets[1]    = socket for control thread communication
                              // sockets[2 ...]= sockets for auxillary communication
                              //
   private int socketCount;    // # of sockets attached to the application
   private int auxSocketCount; // # of auxillary sockets attached to the application

   private Socket startUp;    // For startup communication with the application

   private PrintWriter   controlOut; // Stream wrappers to communicate with the
   private BufferedReader controlIn; // control thread

   boolean verbose;

   PipedInputStream    DataIstream;     // for communcation with local application
   PipedInputStream    ControlIstream;
   PipedOutputStream   DataOstream;
   PipedOutputStream   ControlOstream;

}// class NetworkConnection





